/* $Id: sha1.c 44 2007-01-18 06:11:23Z selinger $ */
/* a reference implementation of the SHA-1 secure hash algorithm,
   defined in FIPS PUB 180-1 */

/* Remarks:
   
   We only hash byte streams, not bit streams, in this implementation.

   We do not reject input stream longer than 2^64 bits. In this
   case, the "length" (for the purpose of padding) is taken mod 2^64.

   Usage: to hash a stream, let s be a pointer to a new sha1_state_t.
   First call sha1_init(s), then call sha1_byte(s, x) for each
   byte x to be hashed, finally call sha1_finish(s). The hash is
   then in s->H[0..4].
*/

#include <string.h>
#include <stdlib.h>

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "sha1.h"

#define f_00(B,C,D) ((D) ^ ((B) & ((C) ^ (D))))      /* bitwise if-then-else */
#define f_20(B,C,D) ((B) ^ (C) ^ (D))
#define f_40(B,C,D) (((B) & (C)) | (((B) | (C)) & (D)))  /* bitwise majority */
#define f_60(B,C,D) ((B) ^ (C) ^ (D))

#define cshift(n,x) (((x) << (n)) | ((x) >> (32-(n)))) /* single instruction */

/* ---------------------------------------------------------------------- */
/* auxiliary */

#ifdef WORDS_BIGENDIAN
# define byteorder 0
# define fix_byteorder(w, n) /* empty */
#else
# define byteorder 3

/* swap byteorder of w[0]..w[n-1] */
static inline void fix_byteorder(uint32_t *w, int n) {
  int i;

  for (i=0; i<n; i++) {
    w[i] = (cshift(8, w[i]) & 0x00ff00ff) | (cshift(24, w[i]) & 0xff00ff00);
  }
}

#endif /* def WORDS_BIGENDIAN */

/* ---------------------------------------------------------------------- */

#ifdef SHA1_REFERENCE

/* apply one "round" of SHA-1 hashing, reference implementation */
static void sha1_process_block_ref(sha1_state_t *s) {
  int t;
  uint32_t A,B,C,D,E;
  uint32_t TEMP;
  uint32_t W[80];

  memcpy(W, s->W, 64);

  for (t=16; t<80; t++) {
    W[t] = W[t-3] ^ W[t-8] ^ W[t-14] ^ W[t-16];
    W[t] = cshift(1, W[t]);
  }

  A = s->H[0];
  B = s->H[1];
  C = s->H[2];
  D = s->H[3];
  E = s->H[4];

  for (t=0; t<20; t++) {
    TEMP = cshift(5, A) + f_00(B,C,D) + E + W[t] + 0x5A827999;
    E = D;
    D = C;
    C = cshift(30, B);
    B = A;
    A = TEMP;
  }

  for (t=20; t<40; t++) {
    TEMP = cshift(5, A) + f_20(B,C,D) + E + W[t] + 0x6Ed9EBA1;
    E = D;
    D = C;
    C = cshift(30, B);
    B = A;
    A = TEMP; 
  }

  for (t=40; t<60; t++) {
    TEMP = cshift(5, A) + f_40(B,C,D) + E + W[t] + 0x8F1BBCDC;
    E = D;
    D = C;
    C = cshift(30, B);
    B = A;
    A = TEMP;
  }

  for (t=60; t<80; t++) {
    TEMP = cshift(5, A) + f_60(B,C,D) + E + W[t] + 0xCA62C1D6;
    E = D;
    D = C;
    C = cshift(30, B);
    B = A;
    A = TEMP;
  }

  s->H[0] += A;
  s->H[1] += B;
  s->H[2] += C;
  s->H[3] += D;
  s->H[4] += E;
}

#endif /* SHA1_REFERENCE */

/* apply one "round" of SHA-1 hashing, fast implementation (on my
   reference i386 system with gcc 3.2.2, this is about 14% faster than
   GNU's sha1sum). */
static void sha1_process_block(sha1_state_t *s) {
  uint32_t A,B,C,D,E;

  A = s->H[0];
  B = s->H[1];
  C = s->H[2];
  D = s->H[3];
  E = s->H[4];

  E += cshift(5, A);
  E += f_00(B,C,D);
  E += s->W[ 0];
  E += 0x5a827999;
  B = cshift(30, B);

  D += cshift(5, E);
  D += f_00(A,B,C);
  D += s->W[ 1];
  D += 0x5a827999;
  A = cshift(30, A);

  C += cshift(5, D);
  C += f_00(E,A,B);
  C += s->W[ 2];
  C += 0x5a827999;
  E = cshift(30, E);

  B += cshift(5, C);
  B += f_00(D,E,A);
  B += s->W[ 3];
  B += 0x5a827999;
  D = cshift(30, D);

  A += cshift(5, B);
  A += f_00(C,D,E);
  A += s->W[ 4];
  A += 0x5a827999;
  C = cshift(30, C);

  E += cshift(5, A);
  E += f_00(B,C,D);
  E += s->W[ 5];
  E += 0x5a827999;
  B = cshift(30, B);

  D += cshift(5, E);
  D += f_00(A,B,C);
  D += s->W[ 6];
  D += 0x5a827999;
  A = cshift(30, A);

  C += cshift(5, D);
  C += f_00(E,A,B);
  C += s->W[ 7];
  C += 0x5a827999;
  E = cshift(30, E);

  B += cshift(5, C);
  B += f_00(D,E,A);
  B += s->W[ 8];
  B += 0x5a827999;
  D = cshift(30, D);

  A += cshift(5, B);
  A += f_00(C,D,E);
  A += s->W[ 9];
  A += 0x5a827999;
  C = cshift(30, C);

  E += cshift(5, A);
  E += f_00(B,C,D);
  E += s->W[10];
  E += 0x5a827999;
  B = cshift(30, B);

  D += cshift(5, E);
  D += f_00(A,B,C);
  D += s->W[11];
  D += 0x5a827999;
  A = cshift(30, A);

  C += cshift(5, D);
  C += f_00(E,A,B);
  C += s->W[12];
  C += 0x5a827999;
  E = cshift(30, E);

  B += cshift(5, C);
  B += f_00(D,E,A);
  B += s->W[13];
  B += 0x5a827999;
  D = cshift(30, D);

  A += cshift(5, B);
  A += f_00(C,D,E);
  A += s->W[14];
  A += 0x5a827999;
  C = cshift(30, C);

  E += cshift(5, A);
  E += f_00(B,C,D);
  E += s->W[15];
  E += 0x5a827999;
  B = cshift(30, B);

  s->W[ 0] ^= s->W[13] ^ s->W[ 8] ^ s->W[ 2];
  s->W[ 0] = cshift(1, s->W[ 0]);
  D += cshift(5, E);
  D += f_00(A,B,C);
  D += s->W[ 0];
  D += 0x5a827999;
  A = cshift(30, A);

  s->W[ 1] ^= s->W[14] ^ s->W[ 9] ^ s->W[ 3];
  s->W[ 1] = cshift(1, s->W[ 1]);
  C += cshift(5, D);
  C += f_00(E,A,B);
  C += s->W[ 1];
  C += 0x5a827999;
  E = cshift(30, E);

  s->W[ 2] ^= s->W[15] ^ s->W[10] ^ s->W[ 4];
  s->W[ 2] = cshift(1, s->W[ 2]);
  B += cshift(5, C);
  B += f_00(D,E,A);
  B += s->W[ 2];
  B += 0x5a827999;
  D = cshift(30, D);

  s->W[ 3] ^= s->W[ 0] ^ s->W[11] ^ s->W[ 5];
  s->W[ 3] = cshift(1, s->W[ 3]);
  A += cshift(5, B);
  A += f_00(C,D,E);
  A += s->W[ 3];
  A += 0x5a827999;
  C = cshift(30, C);

  s->W[ 4] ^= s->W[ 1] ^ s->W[12] ^ s->W[ 6];
  s->W[ 4] = cshift(1, s->W[ 4]);
  E += cshift(5, A);
  E += f_20(B,C,D);
  E += s->W[ 4];
  E += 0x6ed9eba1;
  B = cshift(30, B);

  s->W[ 5] ^= s->W[ 2] ^ s->W[13] ^ s->W[ 7];
  s->W[ 5] = cshift(1, s->W[ 5]);
  D += cshift(5, E);
  D += f_20(A,B,C);
  D += s->W[ 5];
  D += 0x6ed9eba1;
  A = cshift(30, A);

  s->W[ 6] ^= s->W[ 3] ^ s->W[14] ^ s->W[ 8];
  s->W[ 6] = cshift(1, s->W[ 6]);
  C += cshift(5, D);
  C += f_20(E,A,B);
  C += s->W[ 6];
  C += 0x6ed9eba1;
  E = cshift(30, E);

  s->W[ 7] ^= s->W[ 4] ^ s->W[15] ^ s->W[ 9];
  s->W[ 7] = cshift(1, s->W[ 7]);
  B += cshift(5, C);
  B += f_20(D,E,A);
  B += s->W[ 7];
  B += 0x6ed9eba1;
  D = cshift(30, D);

  s->W[ 8] ^= s->W[ 5] ^ s->W[ 0] ^ s->W[10];
  s->W[ 8] = cshift(1, s->W[ 8]);
  A += cshift(5, B);
  A += f_20(C,D,E);
  A += s->W[ 8];
  A += 0x6ed9eba1;
  C = cshift(30, C);

  s->W[ 9] ^= s->W[ 6] ^ s->W[ 1] ^ s->W[11];
  s->W[ 9] = cshift(1, s->W[ 9]);
  E += cshift(5, A);
  E += f_20(B,C,D);
  E += s->W[ 9];
  E += 0x6ed9eba1;
  B = cshift(30, B);

  s->W[10] ^= s->W[ 7] ^ s->W[ 2] ^ s->W[12];
  s->W[10] = cshift(1, s->W[10]);
  D += cshift(5, E);
  D += f_20(A,B,C);
  D += s->W[10];
  D += 0x6ed9eba1;
  A = cshift(30, A);

  s->W[11] ^= s->W[ 8] ^ s->W[ 3] ^ s->W[13];
  s->W[11] = cshift(1, s->W[11]);
  C += cshift(5, D);
  C += f_20(E,A,B);
  C += s->W[11];
  C += 0x6ed9eba1;
  E = cshift(30, E);

  s->W[12] ^= s->W[ 9] ^ s->W[ 4] ^ s->W[14];
  s->W[12] = cshift(1, s->W[12]);
  B += cshift(5, C);
  B += f_20(D,E,A);
  B += s->W[12];
  B += 0x6ed9eba1;
  D = cshift(30, D);

  s->W[13] ^= s->W[10] ^ s->W[ 5] ^ s->W[15];
  s->W[13] = cshift(1, s->W[13]);
  A += cshift(5, B);
  A += f_20(C,D,E);
  A += s->W[13];
  A += 0x6ed9eba1;
  C = cshift(30, C);

  s->W[14] ^= s->W[11] ^ s->W[ 6] ^ s->W[ 0];
  s->W[14] = cshift(1, s->W[14]);
  E += cshift(5, A);
  E += f_20(B,C,D);
  E += s->W[14];
  E += 0x6ed9eba1;
  B = cshift(30, B);

  s->W[15] ^= s->W[12] ^ s->W[ 7] ^ s->W[ 1];
  s->W[15] = cshift(1, s->W[15]);
  D += cshift(5, E);
  D += f_20(A,B,C);
  D += s->W[15];
  D += 0x6ed9eba1;
  A = cshift(30, A);

  s->W[ 0] ^= s->W[13] ^ s->W[ 8] ^ s->W[ 2];
  s->W[ 0] = cshift(1, s->W[ 0]);
  C += cshift(5, D);
  C += f_20(E,A,B);
  C += s->W[ 0];
  C += 0x6ed9eba1;
  E = cshift(30, E);

  s->W[ 1] ^= s->W[14] ^ s->W[ 9] ^ s->W[ 3];
  s->W[ 1] = cshift(1, s->W[ 1]);
  B += cshift(5, C);
  B += f_20(D,E,A);
  B += s->W[ 1];
  B += 0x6ed9eba1;
  D = cshift(30, D);

  s->W[ 2] ^= s->W[15] ^ s->W[10] ^ s->W[ 4];
  s->W[ 2] = cshift(1, s->W[ 2]);
  A += cshift(5, B);
  A += f_20(C,D,E);
  A += s->W[ 2];
  A += 0x6ed9eba1;
  C = cshift(30, C);

  s->W[ 3] ^= s->W[ 0] ^ s->W[11] ^ s->W[ 5];
  s->W[ 3] = cshift(1, s->W[ 3]);
  E += cshift(5, A);
  E += f_20(B,C,D);
  E += s->W[ 3];
  E += 0x6ed9eba1;
  B = cshift(30, B);

  s->W[ 4] ^= s->W[ 1] ^ s->W[12] ^ s->W[ 6];
  s->W[ 4] = cshift(1, s->W[ 4]);
  D += cshift(5, E);
  D += f_20(A,B,C);
  D += s->W[ 4];
  D += 0x6ed9eba1;
  A = cshift(30, A);

  s->W[ 5] ^= s->W[ 2] ^ s->W[13] ^ s->W[ 7];
  s->W[ 5] = cshift(1, s->W[ 5]);
  C += cshift(5, D);
  C += f_20(E,A,B);
  C += s->W[ 5];
  C += 0x6ed9eba1;
  E = cshift(30, E);

  s->W[ 6] ^= s->W[ 3] ^ s->W[14] ^ s->W[ 8];
  s->W[ 6] = cshift(1, s->W[ 6]);
  B += cshift(5, C);
  B += f_20(D,E,A);
  B += s->W[ 6];
  B += 0x6ed9eba1;
  D = cshift(30, D);

  s->W[ 7] ^= s->W[ 4] ^ s->W[15] ^ s->W[ 9];
  s->W[ 7] = cshift(1, s->W[ 7]);
  A += cshift(5, B);
  A += f_20(C,D,E);
  A += s->W[ 7];
  A += 0x6ed9eba1;
  C = cshift(30, C);

  s->W[ 8] ^= s->W[ 5] ^ s->W[ 0] ^ s->W[10];
  s->W[ 8] = cshift(1, s->W[ 8]);
  E += cshift(5, A);
  E += f_40(B,C,D);
  E += s->W[ 8];
  E += 0x8f1bbcdc;
  B = cshift(30, B);

  s->W[ 9] ^= s->W[ 6] ^ s->W[ 1] ^ s->W[11];
  s->W[ 9] = cshift(1, s->W[ 9]);
  D += cshift(5, E);
  D += f_40(A,B,C);
  D += s->W[ 9];
  D += 0x8f1bbcdc;
  A = cshift(30, A);

  s->W[10] ^= s->W[ 7] ^ s->W[ 2] ^ s->W[12];
  s->W[10] = cshift(1, s->W[10]);
  C += cshift(5, D);
  C += f_40(E,A,B);
  C += s->W[10];
  C += 0x8f1bbcdc;
  E = cshift(30, E);

  s->W[11] ^= s->W[ 8] ^ s->W[ 3] ^ s->W[13];
  s->W[11] = cshift(1, s->W[11]);
  B += cshift(5, C);
  B += f_40(D,E,A);
  B += s->W[11];
  B += 0x8f1bbcdc;
  D = cshift(30, D);

  s->W[12] ^= s->W[ 9] ^ s->W[ 4] ^ s->W[14];
  s->W[12] = cshift(1, s->W[12]);
  A += cshift(5, B);
  A += f_40(C,D,E);
  A += s->W[12];
  A += 0x8f1bbcdc;
  C = cshift(30, C);

  s->W[13] ^= s->W[10] ^ s->W[ 5] ^ s->W[15];
  s->W[13] = cshift(1, s->W[13]);
  E += cshift(5, A);
  E += f_40(B,C,D);
  E += s->W[13];
  E += 0x8f1bbcdc;
  B = cshift(30, B);

  s->W[14] ^= s->W[11] ^ s->W[ 6] ^ s->W[ 0];
  s->W[14] = cshift(1, s->W[14]);
  D += cshift(5, E);
  D += f_40(A,B,C);
  D += s->W[14];
  D += 0x8f1bbcdc;
  A = cshift(30, A);

  s->W[15] ^= s->W[12] ^ s->W[ 7] ^ s->W[ 1];
  s->W[15] = cshift(1, s->W[15]);
  C += cshift(5, D);
  C += f_40(E,A,B);
  C += s->W[15];
  C += 0x8f1bbcdc;
  E = cshift(30, E);

  s->W[ 0] ^= s->W[13] ^ s->W[ 8] ^ s->W[ 2];
  s->W[ 0] = cshift(1, s->W[ 0]);
  B += cshift(5, C);
  B += f_40(D,E,A);
  B += s->W[ 0];
  B += 0x8f1bbcdc;
  D = cshift(30, D);

  s->W[ 1] ^= s->W[14] ^ s->W[ 9] ^ s->W[ 3];
  s->W[ 1] = cshift(1, s->W[ 1]);
  A += cshift(5, B);
  A += f_40(C,D,E);
  A += s->W[ 1];
  A += 0x8f1bbcdc;
  C = cshift(30, C);

  s->W[ 2] ^= s->W[15] ^ s->W[10] ^ s->W[ 4];
  s->W[ 2] = cshift(1, s->W[ 2]);
  E += cshift(5, A);
  E += f_40(B,C,D);
  E += s->W[ 2];
  E += 0x8f1bbcdc;
  B = cshift(30, B);

  s->W[ 3] ^= s->W[ 0] ^ s->W[11] ^ s->W[ 5];
  s->W[ 3] = cshift(1, s->W[ 3]);
  D += cshift(5, E);
  D += f_40(A,B,C);
  D += s->W[ 3];
  D += 0x8f1bbcdc;
  A = cshift(30, A);

  s->W[ 4] ^= s->W[ 1] ^ s->W[12] ^ s->W[ 6];
  s->W[ 4] = cshift(1, s->W[ 4]);
  C += cshift(5, D);
  C += f_40(E,A,B);
  C += s->W[ 4];
  C += 0x8f1bbcdc;
  E = cshift(30, E);

  s->W[ 5] ^= s->W[ 2] ^ s->W[13] ^ s->W[ 7];
  s->W[ 5] = cshift(1, s->W[ 5]);
  B += cshift(5, C);
  B += f_40(D,E,A);
  B += s->W[ 5];
  B += 0x8f1bbcdc;
  D = cshift(30, D);

  s->W[ 6] ^= s->W[ 3] ^ s->W[14] ^ s->W[ 8];
  s->W[ 6] = cshift(1, s->W[ 6]);
  A += cshift(5, B);
  A += f_40(C,D,E);
  A += s->W[ 6];
  A += 0x8f1bbcdc;
  C = cshift(30, C);

  s->W[ 7] ^= s->W[ 4] ^ s->W[15] ^ s->W[ 9];
  s->W[ 7] = cshift(1, s->W[ 7]);
  E += cshift(5, A);
  E += f_40(B,C,D);
  E += s->W[ 7];
  E += 0x8f1bbcdc;
  B = cshift(30, B);

  s->W[ 8] ^= s->W[ 5] ^ s->W[ 0] ^ s->W[10];
  s->W[ 8] = cshift(1, s->W[ 8]);
  D += cshift(5, E);
  D += f_40(A,B,C);
  D += s->W[ 8];
  D += 0x8f1bbcdc;
  A = cshift(30, A);

  s->W[ 9] ^= s->W[ 6] ^ s->W[ 1] ^ s->W[11];
  s->W[ 9] = cshift(1, s->W[ 9]);
  C += cshift(5, D);
  C += f_40(E,A,B);
  C += s->W[ 9];
  C += 0x8f1bbcdc;
  E = cshift(30, E);

  s->W[10] ^= s->W[ 7] ^ s->W[ 2] ^ s->W[12];
  s->W[10] = cshift(1, s->W[10]);
  B += cshift(5, C);
  B += f_40(D,E,A);
  B += s->W[10];
  B += 0x8f1bbcdc;
  D = cshift(30, D);

  s->W[11] ^= s->W[ 8] ^ s->W[ 3] ^ s->W[13];
  s->W[11] = cshift(1, s->W[11]);
  A += cshift(5, B);
  A += f_40(C,D,E);
  A += s->W[11];
  A += 0x8f1bbcdc;
  C = cshift(30, C);

  s->W[12] ^= s->W[ 9] ^ s->W[ 4] ^ s->W[14];
  s->W[12] = cshift(1, s->W[12]);
  E += cshift(5, A);
  E += f_60(B,C,D);
  E += s->W[12];
  E += 0xca62c1d6;
  B = cshift(30, B);

  s->W[13] ^= s->W[10] ^ s->W[ 5] ^ s->W[15];
  s->W[13] = cshift(1, s->W[13]);
  D += cshift(5, E);
  D += f_60(A,B,C);
  D += s->W[13];
  D += 0xca62c1d6;
  A = cshift(30, A);

  s->W[14] ^= s->W[11] ^ s->W[ 6] ^ s->W[ 0];
  s->W[14] = cshift(1, s->W[14]);
  C += cshift(5, D);
  C += f_60(E,A,B);
  C += s->W[14];
  C += 0xca62c1d6;
  E = cshift(30, E);

  s->W[15] ^= s->W[12] ^ s->W[ 7] ^ s->W[ 1];
  s->W[15] = cshift(1, s->W[15]);
  B += cshift(5, C);
  B += f_60(D,E,A);
  B += s->W[15];
  B += 0xca62c1d6;
  D = cshift(30, D);

  s->W[ 0] ^= s->W[13] ^ s->W[ 8] ^ s->W[ 2];
  s->W[ 0] = cshift(1, s->W[ 0]);
  A += cshift(5, B);
  A += f_60(C,D,E);
  A += s->W[ 0];
  A += 0xca62c1d6;
  C = cshift(30, C);

  s->W[ 1] ^= s->W[14] ^ s->W[ 9] ^ s->W[ 3];
  s->W[ 1] = cshift(1, s->W[ 1]);
  E += cshift(5, A);
  E += f_60(B,C,D);
  E += s->W[ 1];
  E += 0xca62c1d6;
  B = cshift(30, B);

  s->W[ 2] ^= s->W[15] ^ s->W[10] ^ s->W[ 4];
  s->W[ 2] = cshift(1, s->W[ 2]);
  D += cshift(5, E);
  D += f_60(A,B,C);
  D += s->W[ 2];
  D += 0xca62c1d6;
  A = cshift(30, A);

  s->W[ 3] ^= s->W[ 0] ^ s->W[11] ^ s->W[ 5];
  s->W[ 3] = cshift(1, s->W[ 3]);
  C += cshift(5, D);
  C += f_60(E,A,B);
  C += s->W[ 3];
  C += 0xca62c1d6;
  E = cshift(30, E);

  s->W[ 4] ^= s->W[ 1] ^ s->W[12] ^ s->W[ 6];
  s->W[ 4] = cshift(1, s->W[ 4]);
  B += cshift(5, C);
  B += f_60(D,E,A);
  B += s->W[ 4];
  B += 0xca62c1d6;
  D = cshift(30, D);

  s->W[ 5] ^= s->W[ 2] ^ s->W[13] ^ s->W[ 7];
  s->W[ 5] = cshift(1, s->W[ 5]);
  A += cshift(5, B);
  A += f_60(C,D,E);
  A += s->W[ 5];
  A += 0xca62c1d6;
  C = cshift(30, C);

  s->W[ 6] ^= s->W[ 3] ^ s->W[14] ^ s->W[ 8];
  s->W[ 6] = cshift(1, s->W[ 6]);
  E += cshift(5, A);
  E += f_60(B,C,D);
  E += s->W[ 6];
  E += 0xca62c1d6;
  B = cshift(30, B);

  s->W[ 7] ^= s->W[ 4] ^ s->W[15] ^ s->W[ 9];
  s->W[ 7] = cshift(1, s->W[ 7]);
  D += cshift(5, E);
  D += f_60(A,B,C);
  D += s->W[ 7];
  D += 0xca62c1d6;
  A = cshift(30, A);

  s->W[ 8] ^= s->W[ 5] ^ s->W[ 0] ^ s->W[10];
  s->W[ 8] = cshift(1, s->W[ 8]);
  C += cshift(5, D);
  C += f_60(E,A,B);
  C += s->W[ 8];
  C += 0xca62c1d6;
  E = cshift(30, E);

  s->W[ 9] ^= s->W[ 6] ^ s->W[ 1] ^ s->W[11];
  s->W[ 9] = cshift(1, s->W[ 9]);
  B += cshift(5, C);
  B += f_60(D,E,A);
  B += s->W[ 9];
  B += 0xca62c1d6;
  D = cshift(30, D);

  s->W[10] ^= s->W[ 7] ^ s->W[ 2] ^ s->W[12];
  s->W[10] = cshift(1, s->W[10]);
  A += cshift(5, B);
  A += f_60(C,D,E);
  A += s->W[10];
  A += 0xca62c1d6;
  C = cshift(30, C);

  s->W[11] ^= s->W[ 8] ^ s->W[ 3] ^ s->W[13];
  s->W[11] = cshift(1, s->W[11]);
  E += cshift(5, A);
  E += f_60(B,C,D);
  E += s->W[11];
  E += 0xca62c1d6;
  B = cshift(30, B);

  s->W[12] ^= s->W[ 9] ^ s->W[ 4] ^ s->W[14];
  s->W[12] = cshift(1, s->W[12]);
  D += cshift(5, E);
  D += f_60(A,B,C);
  D += s->W[12];
  D += 0xca62c1d6;
  A = cshift(30, A);

  s->W[13] ^= s->W[10] ^ s->W[ 5] ^ s->W[15];
  s->W[13] = cshift(1, s->W[13]);
  C += cshift(5, D);
  C += f_60(E,A,B);
  C += s->W[13];
  C += 0xca62c1d6;
  E = cshift(30, E);

  s->W[14] ^= s->W[11] ^ s->W[ 6] ^ s->W[ 0];
  s->W[14] = cshift(1, s->W[14]);
  B += cshift(5, C);
  B += f_60(D,E,A);
  B += s->W[14];
  B += 0xca62c1d6;
  D = cshift(30, D);

  s->W[15] ^= s->W[12] ^ s->W[ 7] ^ s->W[ 1];
  s->W[15] = cshift(1, s->W[15]);
  A += cshift(5, B);
  A += f_60(C,D,E);
  A += s->W[15];
  A += 0xca62c1d6;
  C = cshift(30, C);

  s->H[0] += A;
  s->H[1] += B;
  s->H[2] += C;
  s->H[3] += D;
  s->H[4] += E;
}

/* initialize SHA-1 state */
void sha1_init(sha1_state_t *s) {
  s->count_lo = 0;
  s->count_hi = 0;
  s->j = 0;
  s->H[0] = 0x67452301;
  s->H[1] = 0xefcdab89;
  s->H[2] = 0x98badcfe;
  s->H[3] = 0x10325476;
  s->H[4] = 0xc3d2e1f0;
}

/* hash the next byte */
inline void sha1_byte(sha1_state_t *s, int x) {
  unsigned char *p = (unsigned char *)&s->W;  /* access s->W as byte array */

  p[s->j ^ byteorder] = x;
  s->j++;
  s->count_lo+=8;
  if (s->count_lo == 0)
    s->count_hi+=1;
  if (s->j == 16*4) {
    sha1_process_block(s);
    s->j = 0;
  }
}

#ifdef SHA1_REFERENCE

/* hash multiple bytes - reference implementation */
void sha1_bytes_ref(sha1_state_t *s, unsigned char *p, int n) {
  int i;

  for (i=0; i<n; i++) {
    sha1_byte(s, p[i]);
  }
}

#endif /* SHA1_REFERENCE */

/* hash multiple bytes - fast implementation */
void sha1_bytes(sha1_state_t *s, unsigned char *p, int n) {

  /* do the general case, one byte at a time */
  while (n>0 && s->j != 0) {
    sha1_byte(s, *p);
    p++;
    n--;
  }

  /* do the typical case, one block at a time */
  while (n>=64) {
    memcpy(s->W, p, 64);
    fix_byteorder(s->W, 16);
    sha1_process_block(s);
    p+=64;
    n-=64;
    s->count_lo += 512;
    if (s->count_lo == 0) {
      s->count_hi += 1;
    }
  }

  /* finish with general case, one byte at a time */
  while (n>0) {
    sha1_byte(s, *p);
    p++;
    n--;
  }
}

/* Add 0 to 7 bits to the end of the stream, then finish up hashing
   and return final hash. The bits are taken as the n most significant
   bits of the byte buf */
inline void sha1_finish_with_bits(sha1_state_t *s, int n, unsigned char buf) {
  uint32_t count_lo = s->count_lo + n;
  uint32_t count_hi = s->count_hi;

  /* first apply padding */
  buf &= 0xff << (8-n);
  buf |= 0x80 >> n;

  sha1_byte(s, buf);
  while (s->j != 56) {
    sha1_byte(s, 0x00);
  }
  s->W[14] = count_hi;
  s->W[15] = count_lo;

  /* then process last block */
  sha1_process_block(s);
}

/* finish up hashing and return final hash */
void sha1_finish(sha1_state_t *s) {
  sha1_finish_with_bits(s, 0, 0);
}

void sha1_gethash(sha1_state_t *s, unsigned char *p) {
  int i;
  unsigned char *q = (unsigned char *)(s->H);

  for (i=0; i<20; i++) {
    p[i] = q[i ^ byteorder];
  }
}

static void sha1_free(hash_object_t *s) {
  if (s != NULL) {
    free(s->data);
    s->data = NULL;
  }
  free(s);
}

/* return NULL on error with errno set, else an initialized sha1 hash
   object.  */
hash_object_t *sha1_new(void) {
  hash_object_t *self;
  sha1_state_t *data;

  self = (hash_object_t *)malloc(sizeof(hash_object_t));
  if (!self) {
    return NULL;
  }
  data = (sha1_state_t *)malloc(sizeof(sha1_state_t));
  if (!data) {
    free(self);
    return NULL;
  }
  sha1_init(data);

  self->data = data;
  self->hash_len = 20;
  self->hash_init = (void (*)(void *)) sha1_init;
  self->hash_byte = (void (*)(void *, int)) sha1_byte;
  self->hash_bytes = (void (*)(void *, unsigned char *, int)) sha1_bytes;
  self->hash_finish = (void (*)(void *)) sha1_finish;
  self->hash_finish_with_bits = (void (*)(void *, int, unsigned char)) sha1_finish_with_bits;
  self->hash_gethash = (void (*)(void *, unsigned char *)) sha1_gethash;
  self->hash_free = sha1_free;

  return self;
};
